
// Operations on multi-dimensional blocks of data.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "block.h"

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `blockCopy': Untyped version.                                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


void blockCopy(char* dst, int* dstStrs, char* src, int* srcStrs,
               int* exts, int rank, int elementLen) {
  if(rank == 1) {
    const int count   = *exts ;
    const int dstStep = *dstStrs ;
    const int srcStep = *srcStrs ;
    for(int i = 0 ; i < count ; i++) {
      memcpy(dst, src, elementLen) ;
      dst += dstStep ;
      src += srcStep ;
    }
  }
  else {
    rank-- ;
    const int count   = exts [rank] ;
    const int dstStep = elementLen * dstStrs [rank] ;
    const int srcStep = elementLen * srcStrs [rank] ;
    for(int i = 0 ; i < count ; i++) {
      blockCopy(dst, dstStrs, src, srcStrs, exts, rank, elementLen) ;
      dst += dstStep ;
      src += srcStep ;
    }
  }
}
             

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `blockGather': Untyped version.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

char* blockGather(char* dat, int* strs, int* exts, int rank,
                  char* buffer, int elementLen) {
  
  if(rank == 1) {
    const int count   = *exts ;
    const int step    = *strs ;
    const int datStep = elementLen * step ;
    for(int i = 0 ; i < count ; i++) {
      memcpy(buffer, dat, elementLen) ;
      dat    += datStep ;
      buffer += elementLen ;
    }
  }
  else {
    rank-- ;
    const int count   = exts [rank] ;
    const int step    = strs [rank] ;
    const int datStep = elementLen * step ;
    for(int i = 0 ; i < count ; i++) {
      buffer = blockGather(dat, strs, exts, rank, buffer, elementLen) ;
      dat   += datStep ;
    }
  }
  return buffer ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `blockScatter': Untyped version.                                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

char* blockScatter(char* dat, int* strs, int* exts, int rank,
                   char* buffer, int elementLen) {
  
  if(rank == 1) {
    const int count   = *exts ;
    const int step    = *strs ;
    const int datStep = elementLen * step ;
    for(int i = 0 ; i < count ; i++) {
      memcpy(dat, buffer, elementLen) ;
      dat    += datStep ;
      buffer += elementLen ;
    }
  }
  else {
    rank-- ;
    const int count   = exts [rank] ;
    const int step    = strs [rank] ;
    const int datStep = elementLen * step ;
    for(int i = 0 ; i < count ; i++) {
      buffer = blockScatter(dat, strs, exts, rank, buffer, elementLen) ;
      dat   += datStep ;
    }
  }
  return buffer ;
}

// -------------------------------------------------------------
// Things to do
//
//   Should probably cope with rank-0.
//
//   Check for step == 1 in rank == 1 cases of block gather/scatter/copy
//   routines?  Inline code for rank == 1 in calling routines?

